package nicocacheca;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.security.GeneralSecurityException;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Security;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.security.spec.ECGenParameterSpec;
import java.util.Calendar;
import org.bouncycastle.asn1.ASN1EncodableVector;
import org.bouncycastle.asn1.DERSequence;
import org.bouncycastle.asn1.x500.X500Name;
import org.bouncycastle.asn1.x500.X500NameBuilder;
import org.bouncycastle.asn1.x500.style.BCStyle;
import org.bouncycastle.asn1.x509.BasicConstraints;
import org.bouncycastle.asn1.x509.Extension;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.asn1.x509.KeyPurposeId;
import org.bouncycastle.cert.CertIOException;
import org.bouncycastle.cert.X509CertificateHolder;
import org.bouncycastle.cert.X509v3CertificateBuilder;
import org.bouncycastle.cert.jcajce.JcaX509CertificateConverter;
import org.bouncycastle.cert.jcajce.JcaX509CertificateHolder;
import org.bouncycastle.cert.jcajce.JcaX509v3CertificateBuilder;
import org.bouncycastle.jce.X509KeyUsage;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.operator.ContentSigner;
import org.bouncycastle.operator.OperatorCreationException;
import org.bouncycastle.operator.jcajce.JcaContentSignerBuilder;

/**
 * NicoCache_nl+mod+mod用の認証局と証明書を作るよ
 * 
 * 外部ライブラリに依存しているのでNicoCache_nl本体へ
 * 組み込むのを避けて別アプリケーションにしてあります．
 * 
 * Required libraries:
 *  - Bouncy Castle
 *    - bcprov.jar (bcprov-jdk15on-xxx.jar)
 *    - bcpkix.jar (bcpkix-jdk15on-xxx.jar)
 */
public class NicoCacheCA {
    final static String VER_STRING = "NicoCacheCA 170421";
    
    final static String CA_DN = "CN=NicoCache_nl CA";
    final static File CERTS_DIR = new File("certs/");
    final static File CA_KEYSTORE_FILE = new File("certs/ca.jks");
    final static File CA_CERT_FILE = new File("certs/ca.cer");
    final static File SITE_KEYSTORE_FILE = new File("certs/site.jks");
    final static File SITE_CERT_FILE = new File("certs/site.cer");
    final static String KEYSTORE_PASSPHRASE = "NicoCache";
    
    public static boolean generateCA() {
        try {
            KeyPair kp = genKey();
            
            X509v3CertificateBuilder certBuilder = makeDefaultCertificateBuilder(
                    new X500Name(CA_DN),
                    new X500Name(CA_DN),
                    kp.getPublic()
            ).addExtension(
                    Extension.keyUsage,
                    true,
                    new X509KeyUsage(X509KeyUsage.keyCertSign)
            );
            
            X509Certificate cert = caSign(certBuilder, kp.getPrivate());
            
            try (FileOutputStream fos = new FileOutputStream(CA_CERT_FILE)) {
                fos.write(cert.getEncoded());
            }
            
            char[] passphrase = KEYSTORE_PASSPHRASE.toCharArray();
            KeyStore ks = KeyStore.getInstance("JKS");
            ks.load(null, null);
            ks.setKeyEntry("mykey", kp.getPrivate(), passphrase, new Certificate[] { cert });
            ks.store(new FileOutputStream(CA_KEYSTORE_FILE), passphrase);
            
            return true;
        } catch (GeneralSecurityException | OperatorCreationException | IOException ex) {
            ex.printStackTrace();
        }
        return false;
    }
    
    public static boolean generateCertificate(String[] domains) {
        try {
            KeyStore caKeyStore = KeyStore.getInstance("JKS");
            char[] passphrase = KEYSTORE_PASSPHRASE.toCharArray();
            caKeyStore.load(new FileInputStream(CA_KEYSTORE_FILE), passphrase);
            PrivateKey caPrivateKey = (PrivateKey)caKeyStore.getKey("mykey", passphrase);
            X509Certificate caCert = (X509Certificate)caKeyStore.getCertificate("mykey");
            X500Name issuer = new JcaX509CertificateHolder(caCert).getSubject();
            
            KeyPair kp = genKey();
            
            ASN1EncodableVector extendedKeyUsage = new ASN1EncodableVector(); 
            extendedKeyUsage.add(KeyPurposeId.id_kp_serverAuth); 
            extendedKeyUsage.add(KeyPurposeId.id_kp_clientAuth); 
            
            ASN1EncodableVector subjectAltNames = new ASN1EncodableVector();
            for (String domain : domains) {
                subjectAltNames.add(new GeneralName(GeneralName.dNSName, domain));
            }
            
            X500NameBuilder subjectBuilder = new X500NameBuilder(BCStyle.INSTANCE);
            subjectBuilder.addRDN(BCStyle.CN, domains[0]);
            
            X509v3CertificateBuilder certBuilder = makeDefaultCertificateBuilder(
                    issuer,
                    subjectBuilder.build(),
                    kp.getPublic()
            ).addExtension(
                    Extension.keyUsage,
                    true,
                    new X509KeyUsage(X509KeyUsage.digitalSignature)
            ).addExtension(
                    Extension.extendedKeyUsage,
                    false,
                    new DERSequence(extendedKeyUsage)
            ).addExtension(
                    Extension.subjectAlternativeName,
                    false,
                    new DERSequence(subjectAltNames)
            );
            
            X509Certificate cert = caSign(certBuilder, caPrivateKey);
            
            try (FileOutputStream fos = new FileOutputStream(SITE_CERT_FILE)) {
                fos.write(cert.getEncoded());
            }
            
            KeyStore ks = KeyStore.getInstance("JKS");
            ks.load(null, null);
            ks.setKeyEntry("mykey", kp.getPrivate(), passphrase, new Certificate[] { cert });
            ks.store(new FileOutputStream(SITE_KEYSTORE_FILE), passphrase);
            
            return true;
        } catch (GeneralSecurityException | OperatorCreationException | IOException ex) {
            ex.printStackTrace();
        }
        return false;
    }
    
    static KeyPair genKey() throws GeneralSecurityException {
        KeyPairGenerator ecdsa = KeyPairGenerator.getInstance("ECDSA", BouncyCastleProvider.PROVIDER_NAME);
        ecdsa.initialize(new ECGenParameterSpec("prime256v1"));
        KeyPair kp = ecdsa.generateKeyPair();
        return kp;
    }
    
    static X509v3CertificateBuilder makeDefaultCertificateBuilder(X500Name issuer, X500Name subject, PublicKey pk)
            throws CertIOException {
        boolean isCA = issuer.equals(subject);
        
        Calendar notBefore = Calendar.getInstance();
        Calendar notAfter = Calendar.getInstance();
        notAfter.add(Calendar.YEAR, 10);
        
        X509v3CertificateBuilder certBuilder = new JcaX509v3CertificateBuilder(
                issuer,
                isCA ? BigInteger.ONE : BigInteger.valueOf(System.currentTimeMillis()),
                notBefore.getTime(),
                notAfter.getTime(),
                subject,
                pk
        );
        certBuilder.addExtension(
                Extension.basicConstraints,
                true,
                new BasicConstraints(isCA)
        );
        return certBuilder;
    }
    
    static X509Certificate caSign(X509v3CertificateBuilder certBuilder, PrivateKey caPrivateKey)
            throws OperatorCreationException, CertificateException {
        ContentSigner signer =
                new JcaContentSignerBuilder("SHA256withECDSA").build(caPrivateKey);
        X509CertificateHolder certHolder = certBuilder.build(signer);
        X509Certificate cert = new JcaX509CertificateConverter().getCertificate(certHolder);
        return cert;
    }
    
    public static void main(String[] args) {
        if (args.length == 0) {
            System.out.println(VER_STRING);
            System.out.println("Usage: java -jar dist/NicoCacheCA.jar domain ...");
            System.out.println("e.g. \"*.nicovideo.jp\" \"*.ce.nicovideo.jp\" \"*.smilevideo.jp\" ...");
            return;
        }
        if (!CERTS_DIR.exists()) {
            CERTS_DIR.mkdirs();
        }
        Security.addProvider(new BouncyCastleProvider());
        
        boolean isCaGenerated = false;
        if (!CA_KEYSTORE_FILE.exists()) {
            if (!generateCA()) {
                return;
            }
            isCaGenerated = true;
            System.out.println("CA is generated. Please import " + CA_CERT_FILE.getPath() + " to your browser!");
        }
        if (!generateCertificate(args)) {
            return;
        }
        System.out.println("Certificate is generated for " + String.join(", ", args));
        
        if (isCaGenerated) {
            System.out.println("");
            System.out.println("*** Please import " + CA_CERT_FILE.getPath() + " to your browser! ***");
        }
    }
    
}
